@@ -20,11 +20,37 @@ static PyObject *OutOfMemory;
 static PyObject *ConnectionFailure;
 static PyObject *Failure;
 static PyObject *KeyExists;
+static PyObject *AsyncLimit;
 
-#define CB_EXCEPTION(type, msg) { \
-  PyErr_SetString(type, msg); \
-  context->exception = 1; \
-  return 0; \
+#define ASYNC_GUARD() if (context->async_mode) {		\
+    if (context->async_count >= context->async_limit) {		\
+      PyErr_SetString(AsyncLimit, "async limit reached");	\
+      return 0;							\
+    } else if (context->async.count == context->async.size) {	\
+      PyErr_SetString(AsyncLimit, "too many results pending");	\
+      return 0;							\
+    }								\
+  }
+
+#define ASYNC_EXIT(ticket) if(context->async_mode) { \
+    ++context->async_count;			     \
+    return Py_BuildValue("i", ticket[0]);	     \
+  }
+
+#define CB_EXCEPTION(type, msg) {		\
+    PyErr_SetString(type, msg);			\
+    context->exception = 1;			\
+    return 0;					\
+  }
+
+#define NEW_EXCEPTION(type, msg) {					\
+    return PyObject_CallObject(type, Py_BuildValue("(s)", msg));	\
+  }
+
+#define INTERNAL_EXCEPTION_HANDLER(x) {		\
+    if (context->internal_exception == 1) {	\
+      x;					\
+    }						\
   }
 
 /* allocators for commonly held objects */
@@ -96,6 +122,85 @@ void destroy_ticket_slab(struct t_ticket_slab *t) {
   }
 }
 
+/* async structures */
+
+typedef struct t_async_result {
+  int ticket;
+  PyObject *value;
+} async_result;
+
+typedef struct t_async_results {
+  async_result *buffer;
+  int size;
+  int count;
+  int start;
+  int end;
+} async_results;
+
+
+int async_size(async_results *x, int size) {
+  if (size == x->size)
+    return 0;
+
+  async_result *new = malloc(sizeof(async_result) * size);
+  if (!new) {
+    PyErr_SetString(OutOfMemory, "failed to resize async result buffer");
+    return -1;
+  }
+
+  int s = x->start, c = x->count, o = 0;
+  while (c--) {
+    new[o++] = x->buffer[s];
+    s = (s + 1) % x->size;
+  }
+  
+  if (x->buffer)
+    free(x->buffer);
+  x->buffer = new;
+  x->size = size;
+  x->count = o;
+  x->start = 0;
+  x->end = o;
+
+  return 0;
+}
+
+void async_push(async_results *x, int ticket, PyObject *value) {
+  x->count++;
+  x->buffer[x->end].ticket = ticket;
+  x->buffer[x->end].value = value;
+  x->end = (x->end + 1) % x->size;
+}
+
+int process_async_results(async_results *x, int (*f)(async_result *)) {
+  while (x->count) {
+    if (f(&x->buffer[x->start]))
+      return -1;
+    x->start = (x->start + 1) % x->size;
+    x->count--;
+  } return 0;
+}
+
+PyObject *_async_rval;
+
+int begin_async_results() {
+  _async_rval = PyList_New(0);
+  if (!_async_rval) {
+    PyErr_SetString(Failure, "begin_async_results");
+    return -1;
+  } return 0;
+}
+
+int process_async_result(async_result *x) {
+  PyObject *t = PyTuple_Pack(2, PyInt_FromLong(x->ticket), x->value);
+  if (!t) {
+    PyErr_SetString(Failure, "process_async_result");
+    return 0;
+  }
+
+  return PyList_Append(_async_rval, t);
+}
+
 /* structure holding all state for python client */
 
 typedef struct t_pylibcb_instance {
@@ -104,6 +209,10 @@ typedef struct t_pylibcb_instance {
   ticket_slab *ticket_slabs;
   event_list *event_pool;
   event_slab *event_slabs;
+  int async_mode;
+  async_results async;
+  int async_count;
+  int async_limit;
   int succeeded;
   int timed_out;
   int exception;
@@ -228,6 +337,14 @@ int create_timeout(unsigned int usec, int *_ticket) {
 static char *default_error_string = "internal exception";
 static PyObject *default_exception;
 
+PyObject *get_async_results() {
+  if (begin_async_results())
+    return 0;
+  if (process_async_results(&context->async, process_async_result))
+    return 0;
+  return _async_rval;
+}
+
 void *error_callback(libcouchbase_t instance,
 		     libcouchbase_error_t error,
 		     const char *errinfo) {
@@ -242,12 +359,51 @@ void *error_callback(libcouchbase_t instance,
   return 0;
 }
 
-#define INTERNAL_EXCEPTION_HANDLER(x) { \
-  if (context->internal_exception == 1) { \
-    x; \
-  } \
+#define lcb_code(code, type)	  \
+  case code:			  \
+  e_type = type;		  \
+  e_msg = # code;		  \
+  break
+
+#define lcb_fail(code) lcb_code(code, Failure)
+
+PyObject *lcb_error(libcouchbase_error_t err, int context_error) {
+  PyObject *e_type = Failure;
+  char *e_msg = "unknown error passed to lcb_error";
+
+  switch (err) {
+    lcb_fail(LIBCOUCHBASE_SUCCESS);
+    lcb_fail(LIBCOUCHBASE_AUTH_CONTINUE);
+    lcb_fail(LIBCOUCHBASE_AUTH_ERROR);
+    lcb_fail(LIBCOUCHBASE_DELTA_BADVAL);
+    lcb_fail(LIBCOUCHBASE_E2BIG);
+    lcb_fail(LIBCOUCHBASE_EBUSY);
+    lcb_fail(LIBCOUCHBASE_EINTERNAL);
+    lcb_fail(LIBCOUCHBASE_EINVAL);
+    lcb_code(LIBCOUCHBASE_ENOMEM, OutOfMemory);
+    lcb_fail(LIBCOUCHBASE_ERANGE);
+    lcb_fail(LIBCOUCHBASE_ERROR);
+    lcb_fail(LIBCOUCHBASE_ETMPFAIL);
+    lcb_code(LIBCOUCHBASE_KEY_EEXISTS, KeyExists);
+    lcb_fail(LIBCOUCHBASE_KEY_ENOENT);
+    lcb_fail(LIBCOUCHBASE_LIBEVENT_ERROR);
+    lcb_code(LIBCOUCHBASE_NETWORK_ERROR, ConnectionFailure);
+    lcb_fail(LIBCOUCHBASE_NOT_MY_VBUCKET);
+    lcb_fail(LIBCOUCHBASE_NOT_STORED);
+    lcb_fail(LIBCOUCHBASE_NOT_SUPPORTED);
+    lcb_fail(LIBCOUCHBASE_UNKNOWN_COMMAND);
+    lcb_code(LIBCOUCHBASE_UNKNOWN_HOST, ConnectionFailure);
+    lcb_fail(LIBCOUCHBASE_PROTOCOL_ERROR);
+    lcb_code(LIBCOUCHBASE_ETIMEDOUT, Timeout);
+    lcb_code(LIBCOUCHBASE_CONNECT_ERROR, ConnectionFailure);
   }
 
+  if (context_error)
+    CB_EXCEPTION(e_type, e_msg);
+  NEW_EXCEPTION(e_type, e_msg);
+}
+
+
 void *get_callback(libcouchbase_t instance,
 		   const void *cookie,
 		   libcouchbase_error_t error,
@@ -257,8 +413,30 @@ void *get_callback(libcouchbase_t instance,
 		   libcouchbase_size_t nbytes,
 		   libcouchbase_uint32_t flags,
 		   libcouchbase_cas_t cas) {
+  int t = rip_ticket((int *) cookie);
+  if (context->async_mode) {
+    PyObject *rval;
+    --context->async_count;
 
-  if (rip_ticket((int *) cookie) != context->callback_ticket)
+    switch (error) {
+    case LIBCOUCHBASE_SUCCESS:
+      rval = Py_BuildValue("(s#k)", bytes, nbytes, (unsigned long) cas);
+      break;
+
+    case LIBCOUCHBASE_KEY_ENOENT:
+      Py_INCREF(Py_None);
+      rval = Py_None;
+      break;
+      
+    default:
+      rval = lcb_error(error, 0);
+    }
+
+    async_push(&context->async, t, rval);
+    return 0;
+  }
+
+  if (t != context->callback_ticket)
     return 0;
   context->result = error;
 
@@ -268,12 +446,8 @@ void *get_callback(libcouchbase_t instance,
   case LIBCOUCHBASE_KEY_ENOENT:
     context->succeeded = 1;
     return 0;
-  case LIBCOUCHBASE_ENOMEM:
-    CB_EXCEPTION(OutOfMemory, "libcouchbase_enomem");
-  case LIBCOUCHBASE_EINTERNAL:
-    CB_EXCEPTION(Failure, "libcouchbase_einternal");
   default:
-    CB_EXCEPTION(Failure, "mystery condition in get callback");
+    return lcb_error(error, 1);
   }
   
   context->returned_value = Py_BuildValue("s#", bytes, nbytes);
@@ -290,17 +464,33 @@ void *set_callback(libcouchbase_t instance,
 		   const void *key,
 		   libcouchbase_size_t nkey,
 		   libcouchbase_cas_t cas) {
-  if (rip_ticket((int *) cookie) != context->callback_ticket)
+  int t = rip_ticket((int *) cookie);
+  if (context->async_mode) {
+    PyObject *rval;
+    --context->async_count;
+    
+    switch(error) {
+    case LIBCOUCHBASE_SUCCESS:
+      Py_INCREF(Py_True);
+      rval = Py_True;
+      break;
+    default:
+      rval = lcb_error(error, 0);
+    }
+
+    async_push(&context->async, t, rval);
+    return 0;
+  }
+
+  if (t != context->callback_ticket)
     return 0;
   context->result = error;
 
   switch (error) {
   case LIBCOUCHBASE_SUCCESS:
     break;
-  case LIBCOUCHBASE_KEY_EEXISTS:
-    CB_EXCEPTION(KeyExists, "key already present");
   default:
-    CB_EXCEPTION(Failure, "mystery condition in set callback");
+    return lcb_error(error, 1);
   }
 
   return 0;
@@ -311,7 +501,25 @@ void *remove_callback(libcouchbase_t instance,
 		      libcouchbase_error_t error,
 		      const void *key,
 		      libcouchbase_size_t nkey) {
-  if (rip_ticket((int *) cookie) != context->callback_ticket)
+  int t = rip_ticket((int *) cookie);
+  if (context->async_mode) {
+    PyObject *rval;
+    --context->async_count;
+    
+    switch (error) {
+    case LIBCOUCHBASE_SUCCESS:
+      Py_INCREF(Py_True);
+      rval = Py_True;
+      break;
+    default:
+      rval = lcb_error(error, 0);
+    }
+
+    async_push(&context->async, t, rval);
+    return 0; 
+  }
+
+  if (t != context->callback_ticket)
     return 0;
   context->result = error;
 
@@ -319,7 +527,7 @@ void *remove_callback(libcouchbase_t instance,
   case LIBCOUCHBASE_SUCCESS:
     break;
   default:
-    CB_EXCEPTION(Failure, "mystery condition in set callback");
+    return lcb_error(error, 1);
   }  
 
   return 0;
@@ -408,6 +616,8 @@ static PyObject *open(PyObject *self, PyObject *args) {
   default_error_string = "internal exception";
   default_exception = Failure;
 
+  z->async_limit = 20; /* maximum number of async events that may be queued up on the event loop */
+
   return PyCObject_FromVoidPtrAndDesc(z, pylibcb_instance_desc, pylibcb_instance_dest);
 
  free_event_base:
@@ -448,6 +658,77 @@ int *new_ticket() {
   } return t;
 }
 
+static PyObject *get_async_limit(PyObject *self, PyObject *args) {
+  PyObject *cb;
+  
+  if (!PyArg_ParseTuple(args, "O", &cb))
+    return 0;
+
+  set_context(cb);
+
+  return PyInt_FromLong(context->async_limit);
+}
+
+static PyObject *set_async_limit(PyObject *self, PyObject *args) {
+  PyObject *cb;
+  int limit;
+
+  if (!PyArg_ParseTuple(args, "Oi", &cb, &limit))
+    return 0;
+
+  if (limit < 1) {
+    PyErr_SetString(Failure, "limit must be an integer value greater than 0");
+    return 0;
+  } else if (limit > 16384) {
+    PyErr_SetString(Failure, "there is an arbitrary upper limit of 16384 simultaneous requests as that is enough for anyone");
+    return 0;
+  }
+
+  set_context(cb);
+
+  if (context->async.count * 2 < limit)
+    async_size(&context->async, limit * 2);
+  context->async_limit = limit;
+
+  Py_RETURN_NONE;
+}
+
+static PyObject *get_async_count(PyObject *self, PyObject *args) {
+  PyObject *cb;
+
+  if (!PyArg_ParseTuple(args, "O", &cb))
+    return 0;
+
+  set_context(cb);
+  return PyInt_FromLong(context->async_count);
+}
+
+static PyObject *enable_async(PyObject *self, PyObject *args) {
+  PyObject *cb;
+
+  if (!PyArg_ParseTuple(args, "O", &cb))
+      return 0;
+  set_context(cb);
+
+  context->async_mode = 1;
+  context->async_count = 0;
+  if (context->async.count < context->async_limit * 2)
+    async_size(&context->async, context->async_limit * 2);
+
+  Py_RETURN_NONE;
+}
+
+static PyObject *disable_async(PyObject *self, PyObject *args) {
+  PyObject *cb;
+
+  if (!PyArg_ParseTuple(args, "O", &cb))
+      return 0;
+  set_context(cb);
+
+  context->async_mode = 0;
+
+  Py_RETURN_NONE;
+}
 
 static PyObject *set(PyObject *self, PyObject *args) {
   PyObject *cb;
@@ -458,10 +739,10 @@ static PyObject *set(PyObject *self, PyObject *args) {
 
   if (!PyArg_ParseTuple(args, "Os#s#|kk", &cb, &key, &nkey, &val, &nval, &_expiry, &cas))
     return 0;
-  if (!pyobject_is_pylibcb_instance(cb))
-    return 0;
   set_context(cb);
 
+  ASYNC_GUARD();
+
   time_t expiry = _expiry;
   int *ticket = new_ticket();
   if (!ticket)
@@ -469,14 +750,15 @@ static PyObject *set(PyObject *self, PyObject *args) {
 
   libcouchbase_store_by_key(context->cb, hand_out_ticket(ticket), LIBCOUCHBASE_SET, 0, 0, 
 			    key, nkey, val, nval, 0, expiry, cas);
+  ASYNC_EXIT(ticket);
+
   libcouchbase_wait(context->cb);
   INTERNAL_EXCEPTION_HANDLER(return 0);
 
   if (context->exception)
     return 0;
 
-  Py_INCREF(Py_None);
-  return Py_None;
+  Py_RETURN_NONE;
 }
 
 static PyObject *_remove(PyObject *self, PyObject *args) {
@@ -487,23 +769,24 @@ static PyObject *_remove(PyObject *self, PyObject *args) {
 
   if (!PyArg_ParseTuple(args, "Os#|k", &cb, &key, &nkey, &cas))
     return 0;
-  if (!pyobject_is_pylibcb_instance(cb))
-    return 0;
   set_context(cb);
 
+  ASYNC_GUARD();
+
   int *ticket = new_ticket();
   if (!ticket)
     return 0;
 
   libcouchbase_remove_by_key(context->cb, hand_out_ticket(ticket), 0, 0, key, nkey, cas);
+  ASYNC_EXIT(ticket);
+
   libcouchbase_wait(context->cb);
   INTERNAL_EXCEPTION_HANDLER(return 0);
 
   if (context->exception)
     return 0;
 
-  Py_INCREF(Py_None);
-  return Py_None;
+  Py_RETURN_NONE;
 }
 
 static PyObject *get(PyObject *self, PyObject *args) {
@@ -516,22 +799,24 @@ static PyObject *get(PyObject *self, PyObject *args) {
   
   if (!PyArg_ParseTuple(args, "Os#|iki", &cb, &key, &_nkey, &usec, &_expiry, &return_cas))
     return 0;
-  if (!pyobject_is_pylibcb_instance(cb))
-    return 0;
   set_context(cb);
 
+  ASYNC_GUARD();
+
   libcouchbase_size_t nkey = _nkey;
   time_t expiry = _expiry;
   int *ticket = new_ticket();
   if (!ticket)
     return 0;
 
-  if (usec)
+  if (usec && !context->async_mode)
     if (!create_timeout(usec, hand_out_ticket(ticket))) {
       rip_ticket(ticket);
       return 0;
     }
+
   libcouchbase_mget_by_key(context->cb, hand_out_ticket(ticket), 0, 0, 1, &key, &nkey, _expiry ? &expiry : 0);
+  ASYNC_EXIT(ticket);
 
   while (!context->timed_out && !context->succeeded && !context->exception && !context->internal_exception)
     libcouchbase_wait(context->cb);
@@ -550,16 +835,50 @@ static PyObject *get(PyObject *self, PyObject *args) {
     return 0;
   }
 
-  if (context->result == LIBCOUCHBASE_KEY_ENOENT) {
-    Py_INCREF(Py_None);
-    return Py_None;
-  }
+  if (context->result == LIBCOUCHBASE_KEY_ENOENT)
+    Py_RETURN_NONE;
 
   if (return_cas)
     return Py_BuildValue("Ok", context->returned_value, (unsigned long) context->returned_cas);
   return context->returned_value;
 }
 
+static PyObject *async_wait(PyObject *self, PyObject *args) {
+  PyObject *cb;
+  int usec = 0;
+
+  if (!PyArg_ParseTuple(args, "Oi", &cb, &usec))
+    return 0;
+  set_context(cb);
+
+  if (!context->async_mode) {
+    PyErr_SetString(Failure, "async mode is not enabled");
+    return 0;
+  }
+
+  if (usec) {
+    int *ticket = new_ticket();
+    if (!ticket)
+      return 0;
+
+    if (!create_timeout(usec, hand_out_ticket(ticket))) {
+      rip_ticket(ticket);
+      return 0;
+    }
+  }
+
+  while (!context->timed_out && context->async_count && !context->internal_exception)
+    libcouchbase_wait(context->cb);
+
+  INTERNAL_EXCEPTION_HANDLER(return 0);
+
+  PyObject *r = get_async_results();
+  if (!r) {
+    PyErr_SetString(Failure, "get_async_results");
+    return 0;
+  } return r;
+}
+
 static PyMethodDef PylibcbMethods[] = {
   { "open", open, METH_VARARGS,
     "Open connection to couchbase server" },
@@ -569,6 +888,18 @@ static PyMethodDef PylibcbMethods[] = {
     "Set a value by key" },
   { "remove", _remove, METH_VARARGS,
     "Remove a value by key" },
+  { "get_async_limit", get_async_limit, METH_VARARGS,
+    "Get the limit for the number of requests allowed before one is required to complete" },
+  { "set_async_limit", set_async_limit, METH_VARARGS,
+    "Set the limit for the number of requests allowed before one is required to complete" },
+  { "get_async_count", get_async_count, METH_VARARGS,
+    "Get the number of incomplete asynchronous requests waiting" },
+  { "enable_async", enable_async, METH_VARARGS,
+    "Enable asynchronous behavior" },
+  { "disable_async", disable_async, METH_VARARGS,
+    "Disable asynchronous behavior" },
+  { "async_wait", async_wait, METH_VARARGS,
+    "Execute eventloop for a given number of microseconds" },
   { 0, 0, 0, 0 }
 };
 
@@ -578,26 +909,29 @@ PyMODINIT_FUNC init_pylibcb() {
   m = Py_InitModule("_pylibcb", PylibcbMethods);
   if (!m)
     return;
-
-  Timeout = PyErr_NewException("_pylibcb.Timeout", 0, 0);
-  Py_INCREF(Timeout);
-  PyModule_AddObject(m, "Timeout", Timeout);
   
-  OutOfMemory = PyErr_NewException("_pylibcb.OutOfMemory", 0, 0);
-  Py_INCREF(OutOfMemory);
-  PyModule_AddObject(m, "OutOfMemory", OutOfMemory);
+  struct exception_init {
+    char *name;
+    PyObject **exception;
+  } exceptions[] = {
+    { "Timeout", &Timeout },
+    { "OutOfMemory", &OutOfMemory },
+    { "ConnectionFailure", &ConnectionFailure },
+    { "Failure", &Failure },
+    { "KeyExists", &KeyExists },
+    { "AsyncLimit", &AsyncLimit },
+    { 0, 0 }
+  };
 
-  ConnectionFailure = PyErr_NewException("_pylibcb.ConnectionFailure", 0, 0);
-  Py_INCREF(ConnectionFailure);
-  PyModule_AddObject(m, "ConnectionFailure", ConnectionFailure);
-  
-  Failure = PyErr_NewException("_pylibcb.Failure", 0, 0);
-  Py_INCREF(Failure);
-  PyModule_AddObject(m, "Failure", Failure);
-
-  KeyExists = PyErr_NewException("_pylibcb.KeyExists", 0, 0);
-  Py_INCREF(KeyExists);
-  PyModule_AddObject(m, "KeyExists", KeyExists);
+  int i = 0;
+  while (exceptions[i].name) {
+    char name[256];
+    sprintf(name, "_pylibcb.%s", exceptions[i].name);
+    *exceptions[i].exception = PyErr_NewException(name, 0, 0);
+    Py_INCREF(*exceptions[i].exception);
+    PyModule_AddObject(m, exceptions[i].name, *exceptions[i].exception);
+    ++i;
+  }
 
   default_exception = Failure;
 }
